import json
import os
import re
import difflib
from functools import lru_cache

import edlib
import unidecode

from oireachtas_data.constants import MEMBERS_DIR


class Members():

    def __init__(self):
        self.loaded = False
        self.data = []

    def append(self, member):
        self.data.append(member)

    def serialize(self):
        return [
            m.serialize() for m in self.data
        ]

    def write(self):
        with open(os.path.join(MEMBERS_DIR, 'members.json'), 'w') as f:
            f.write(
                json.dumps(self.serialize())
            )

    def load(self):
        if self.loaded:
            return

        if os.path.exists(os.path.join(MEMBERS_DIR, 'members.json')):
            with open(os.path.join(MEMBERS_DIR, 'members.json'), 'r') as f:
                data = json.loads(f.read())
                for d in data:
                    self.data.append(
                        Member(**d)
                    )
        self.loaded = True

    def get_member(self, member_str):
        if member_str.startswith('#'):
            return self.get_member_from_id(member_str)
        else:
            return self.get_member_from_name(member_str)

    @lru_cache(maxsize=500)
    def get_member_from_name(self, name):

        # If a minister is specified it usually looks like:
        # Minister for Something (Deputy Joan Malone)
        if 'Minister' in name:
            try:
                name = re.findall('\((.*?)\)', name)[0]
            except:
                pass

        name = name.replace('Deputy ', '')
        name = name.replace('Senator ', '')
        name = name.replace('Mr. ', '')
        name = name.replace('Ms. ', '')
        name = name.replace('Mrs. ', '')
        name = name.replace(' ', '')

        # Remove fadas
        name = unidecode.unidecode(name)

        name = name.replace('\'', '')

        for member in self.data:
            if member.pid is None:
                continue
            if member.pid == name:
                return member

            # see if we can determine if they're too far apart so we can exit early. Can do this stupidly

            if len(list(set(member.pid).symmetric_difference(name))) > 2:
                continue

            # FIXME: super slow
            # Could keep a set of members we got and skip over them since this will only be reached once cause of the cache. Bit ugly
            if edlib.align(member.pid, name)['editDistance'] < 4:
                return member

    @lru_cache(maxsize=500)
    def get_member_from_id(self, pid):
        pid = pid.replace('#', '')

        for member in self.data:
            if member.pid == pid:
                return member

    @lru_cache(maxsize=500)
    def parties_of_member(self, member):

        if isinstance(member, str):
            member = self.get_member_from_name(member)
            if member is None:
                return None

        parties = []
        for membership in member.memberships:
            parties.append(membership.parties[0].party_code)

        return parties

class Member():

    def __init__(self, *args, **kwargs):
        self.date_of_death = kwargs.get('dateOfDeath', kwargs.get('date_of_death', None))
        self.first_name = kwargs.get('firstName', kwargs.get('first_name', None))
        self.last_name = kwargs.get('lastName', kwargs.get('last_name', None))
        self.full_name = kwargs.get('fullName', kwargs.get('full_name', None))
        self.gender = kwargs.get('gender', kwargs.get('gender', None))
        self.member_code = kwargs.get('memberCode', kwargs.get('member_code', None))
        self.pid = kwargs.get('pId', kwargs.get('pid', None))

        self.memberships = []
        for membership in kwargs['memberships']:
            if 'membership' in membership:
                self.memberships.append(
                    Membership(**membership['membership'])
                )
            else:
                self.memberships.append(
                    Membership(**membership)
                )

    def serialize(self):
        return {
            'date_of_death': self.date_of_death,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'full_name': self.full_name,
            'gender': self.gender,
            'member_code': self.member_code,
            'pid': self.pid,
            'memberships': [
                m.serialize() for m in self.memberships
            ]

        }


class House():

    def __init__(self, *args, **kwargs):
        self.chamber_type = kwargs.get('chamberType', kwargs.get('chamber_type', None))
        self.house_code = kwargs.get('houseCode', kwargs.get('house_code', None))
        self.house_no = kwargs.get('houseNo', kwargs.get('house_no', None))
        self.show_as = kwargs.get('showAs', kwargs.get('show_as', None))

    def serialize(self):
        return {
            'chamber_type': self.chamber_type,
            'house_code': self.house_code,
            'house_no': self.house_no,
            'show_as': self.show_as
        }


class Party():

    def __init__(self, *args, **kwargs):

        if 'party' in kwargs:
            base = kwargs['party']

            date_range = base['dateRange']

            self.date_start = date_range['start']
            self.date_end = date_range['end']

            self.party_code = base['partyCode']
            self.show_as = base['showAs']

        else:

            self.date_start = kwargs['date_start']
            self.date_end = kwargs['date_end']

            self.party_code = kwargs['party_code']
            self.show_as = kwargs['show_as']

    def serialize(self):
        return {
            'date_start': self.date_start,
            'date_end': self.date_end,
            'party_code': self.party_code,
            'show_as': self.show_as
        }


class Represents():

    def __init__(self, *args, **kwargs):
        if 'represent' in kwargs:
            base = kwargs['represent']

            self.represent_code = base['representCode']
            self.represent_type = base['representType']
            self.show_as = base['showAs']

        else:

            self.represent_code = kwargs['represent_code']
            self.represent_type = kwargs['represent_type']
            self.show_as = kwargs['show_as']

    def serialize(self):
        return {
            'represent_code': self.represent_code,
            'represent_type': self.represent_type,
            'show_as': self.show_as
        }


class Membership():

    def __init__(self, *args, **kwargs):

        if 'dateRange' in kwargs:
            date_range = kwargs['dateRange']
            self.date_start = date_range.get('start', None)
            self.date_end = date_range.get('end', None)
        else:
            self.date_start = kwargs['date_start']
            self.date_end = kwargs['date_end']

        self.house = House(**kwargs['house'])

        self.offices = []
        for office in kwargs['offices']:
            pass  # TODO

        self.parties = []
        for party in kwargs['parties']:
            self.parties.append(
                Party(
                    **party
                )
            )

        self.represents = []
        for rep in kwargs['represents']:
            self.represents.append(
                Represents(
                    **rep
                )
            )

    def serialize(self):
        return {
            'date_start': self.date_start,
            'date_end': self.date_end,
            'house': self.house.serialize(),
            'offices': [o.serialize() for o in self.offices],
            'parties': [p.serialize() for p in self.parties],
            'represents': [r.serialize() for r in self.represents]
        }



import os

import requests

from oireachtas_data.models.member import Members, Member
from oireachtas_data.constants import MEMBERS_DIR


def main():

    if not os.path.exists(MEMBERS_DIR):
        os.makedirs(MEMBERS_DIR, exist_ok=True)

    chambers = [
        'dail',
        'seanad'
    ]

    base_url = "https://api.oireachtas.ie/v1/members?date_start=1900-01-01&chamber_id=&chamber={chamber}&date_end=2099-01-01&limit=9999"

    members = Members()

    for chamber in chambers:
        url = base_url.format(chamber=chamber)

        result = requests.get(url)

        for result in result.json()['results']:
            member = Member(**result['member'])
            members.append(member)

    members.write()


if __name__ == '__main__':
    main()



import os
import datetime
from difflib import SequenceMatcher
from collections import defaultdict

from cached_property import cached_property
import bs4

from oireachtas_data.models.speech import Speech
from oireachtas_data.models.para import Para


class Section():

    def __init__(self, data=None, title=None, date_string=None):
        self.date_string = date_string
        self.data = data
        self.title = title

    @property
    def content(self):
        split = self.data.split('\n')
        count = 0
        while self.date_string in split:
            count += 1
            if count > 1000:
                break
            idx = split.index(self.date_string)

            if idx == 0:
                split = split[1:]
            elif idx == 1:
                split = split[2:]
            elif idx == 2:
                split = split[3:]
            elif idx == 3:
                split = split[4:]
            else:

                prev2 = split[idx - 2]

                if prev2 != '':
                    continue

                # TODO: check that idx + 2 is always correct
                split = split[0:idx-4] + split[idx + 2:]

        # TODO: need to join broken sentences
        # TODO: need to remove [NAME] casue that's a page break

        final = []

        for line in split:
            if ':' in line:
                final.append(line)
            else:
                if len(final):
                    final[-1] = final[-1] + ' ' + line
                else:
                    pass  # First line has no : so possibly the question

        return '\n'.join(final)

    def non_header_content(self):
        return ''

    @cached_property
    def speeches(self):

        # FIXME if questions it goes '{int}. {NAME} the question'
        if ':' not in self.content:
            return []

        speeches = []

        lines = self.content.split('\n')
        for line in lines:
            if ':' not in line:
                continue

            person = line[:line.index(':')]
            content = line[line.index(':') + 2:]

            speeches.append(
                Speech(
                    by=person,
                    _as=None,
                    eid=None,
                    paras=[
                        Para(
                            title=None,
                            eid=None,
                            content=content
                        )
                    ]
                )
            )

        return speeches


class Page():

    def __init__(self, data, date_string):
        self.data = data
        self.date_string = date_string

    @property
    def header(self):
        from oireachtas_data.utils import find_nth
        return self.data[
            0:find_nth(self.data, '\n\n', 3)
        ].replace('\n\n' + self.date_string + '\n\n', ' ').strip()

    @property
    def sans_header(self):
        from oireachtas_data.utils import find_nth
        return self.data[find_nth(self.data, '\n\n', 3):].strip()

    def serialize(self):
        return {
            'data': self.data,
            'date_string': self.date_string,
            'sans_header': self.sans_header,
            'header': self.header
        }


class PDF():

    def __init__(self, fp):
        self.fp = fp
        self.data = None
        self.loaded = False
        self.removed_sections = []

    def load(self):
        if self.loaded:
            return

        html_file = self.fp.replace('.pdf', 's.html')

        if not os.path.exists(html_file):
            os.system(f'pdftohtml \'{self.fp}\'')

        if not os.path.exists(html_file):
            print('Could not convert pdf to html: %s' % (html_file))
            self.loaded = True
            return

        # TODO: parse xml tree
        soup = bs4.BeautifulSoup(open(html_file), 'html.parser')

        bolds = soup.find_all('b')

        possibles = []
        for bold in bolds:
            if hasattr(bold.nextSibling, 'name'):
                if bold.nextSibling.name != 'br':
                    continue

            # Is it even valid excluding these?
            if 'Deputy' in bold.text:
                continue
            if 'Minister' in bold.text:
                continue
            if 'Senator' in bold.text:
                continue

            ## Maybe exclude the following?
            #if 'Comhairle' in bold.text:
            #    continue
            #if 'Cathaoirleach' in bold.text:
            #    continue
            #if 'Taoiseach' in bold.text:
            #    continue

            # Could clean up more but not much of a point

            possibles.append(bold)

        self.possibles = possibles

        self.loaded = True

    def matching_header(self, show_as):

        self.load()

        space = '\xa0'

        # TODO: factor in removed ones

        for header in self.possibles:
            if SequenceMatcher(None, header.text.replace(space, ' '), show_as).ratio() > 0.6:
                return header.text.replace(space, ' ')

        return None

    @property
    def section_headers(self):
        if not self.loaded:
            self.load()

        if self.data is None:
            return []

        return []

    @property
    def date(self):
        return datetime.datetime.strptime(
            os.path.splitext(self.fp.split('_')[-1])[0],
            '%Y-%m-%d'
        )

    @property
    def date_string(self):
        return self.date.strftime('%-d %B %Y.')

    @property
    def debate_sections(self):
        return self._debate_sections

    @cached_property
    def _debate_sections(self):
        if not self.loaded:
            self.load()

        sections = []

        html_file = self.fp.replace('.pdf', 's.html')
        html_lines = []
        with open(html_file, 'r') as f:
            html_lines = f.readlines()

        header_line_no_map = defaultdict(list)
        line_no_header_map = {}

        last_index = 0
        for possible in self.possibles:
            for idx, line in enumerate(html_lines[last_index:]):
                # can skip a lot of stuff before parsing
                if bs4.BeautifulSoup(line, 'html.parser').text.strip() == possible.text.strip():
                    # go until the next possible and then set last index
                    last_index += idx + 1
                    possible_text = possible.text
                    header_line_no_map[possible_text.strip().replace('\xa0', ' ')].append(last_index)
                    line_no_header_map[last_index] = possible_text.strip().replace('\xa0', ' ')
                    break

        from oireachtas_data.utils import window
        for x, y in window(line_no_header_map.items(), window_size=2):
            content = html_lines[x[0]:y[0]-1]

            lines_text = [
                bs4.BeautifulSoup(c.replace('<br/>', '\n'), 'html.parser').text.strip().replace('\xa0', ' ') for c in content
            ]

            # TODO: in content need to make <br/> be a \n or else it joins directly into the next word
            sections.append(
                Section(
                    date_string=self.date_string,
                    data='\n'.join(lines_text),
                    title=x[1]
                )
            )

        for s in sections:
            s.speeches

        return sections



import os
import datetime
import json
import requests
from collections import defaultdict

from oireachtas_data.constants import DEBATES_DIR
from oireachtas_data.models.debate_section import DebateSection


class Debate():

    __slots__ = (
        'date',
        'chamber',
        'counts',
        'debate_sections',
        'debate_type',
        'data_uri',
        'loaded',
        '_json_location'
    )

    def __init__(
        self,
        date=None,
        chamber=None,
        counts=None,
        debate_sections=None,
        debate_type=None,
        data_uri=None,
        loaded=False
    ):
        '''

        :kwarg date: The date of the debate in YYYY-MM-DD format
        :kwarg chamber: Dail or Seanad
        :kwarg counts:
        :kwarg debate_sections: list of debate sections in json from the api
        :kwarg debate_type: the tye of the debate, usually just 'debate'
        :kwarg data_uri:
        '''
        self.date = date
        self.chamber = chamber
        self.counts = counts
        self.debate_sections = debate_sections
        self.debate_type = debate_type
        self.data_uri = data_uri

        self.loaded = loaded

        self._json_location = None

    def set_filepath(self, filepath):
        self._json_location = filepath

    @staticmethod
    def from_file(filepath):
        from oireachtas_data.utils import get_file_content
        debate = Debate()
        debate.set_filepath(filepath)
        return debate

    @staticmethod
    def parse(data):
        return Debate(
            date=data['date'],
            chamber=data['chamber'],
            counts=data['counts'],
            debate_sections=[
                DebateSection.parse(d) for d in data['sections']
            ],
            debate_type=data['debate_type'],
            data_uri=data['data_uri'],
            loaded=True
        )

    def load_data(self):
        '''
        Load data from the data_uri to populate the debate sections
        '''
        if self.loaded:
            return

        if os.path.exists(self.json_location):
            from oireachtas_data.utils import get_file_content
            data = get_file_content(self.json_location)
            self.date = data['date']
            self.chamber = data['chamber']
            self.counts = data['counts']
            self.debate_sections = [
                DebateSection.parse(d) for d in data['sections']
            ]
            self.debate_type = data['debate_type']
            self.data_uri = data['data_uri']
        else:

            chambers = {
                'Dáil Éireann': 'dail',
                'Seanad Éireann': 'seanad'
            }
            url = 'https://data.oireachtas.ie/ie/oireachtas/debateRecord/%s/%s/debate/mul@/main.pdf' % (
                chambers[self.chamber],
                self.date
            )

            pdf_request = requests.get(url, stream=True)

            # not self.pdf_location as recursive error
            pdf_location = os.path.join(
                DEBATES_DIR,
                '%s_%s_%s.pdf' % ('debate', self.chamber, self.date)
            )

            with open(pdf_location, 'wb') as pdfile:
                for chunk in pdf_request.iter_content(2000):
                    pdfile.write(chunk)

            # If we're missing data from the api cause of being forbidden we can get some data from the pdf

            debate_sections = []
            for idx, section in enumerate(self.debate_sections):
                raw_debate_section = section['debateSection']
                debate_section = DebateSection(
                    bill=raw_debate_section['bill'],
                    contains_debate=raw_debate_section['containsDebate'],
                    counts=raw_debate_section['counts'],
                    debate_section_id=raw_debate_section['debateSectionId'],
                    debate_type=raw_debate_section['debateType'],
                    data_uri=raw_debate_section['formats']['xml']['uri'],
                    parent_debate_section=raw_debate_section['parentDebateSection'],
                    show_as=raw_debate_section['showAs'],
                    show_as_idx=idx,
                    show_as_context=[d['debateSection']['showAs'] for d in self.debate_sections],
                    speakers=raw_debate_section['speakers'],
                    pdf_location=pdf_location
                )
                debate_section.load_data()
                debate_sections.append(debate_section)

            self.debate_sections = debate_sections

        self.loaded = True

    def serialize(self):
        self.load_data()
        return {
            'date': self.date,
            'chamber': self.chamber,
            'counts': self.counts,
            'debate_type': self.debate_type,
            'data_uri': self.data_uri,
            'sections': [
                s.serialize() for s in self.debate_sections
            ]
        }

    def write(self):
        with open(self.json_location, 'w') as outfile:
            outfile.write(json.dumps(self.serialize()))

    @property
    def pdf_location(self):
        self.load_data()
        return os.path.join(
            DEBATES_DIR,
            '%s_%s_%s.pdf' % ('debate', self.chamber, self.date)
        )

    @property
    def json_location(self):
        if self._json_location:
            return self._json_location

        return os.path.join(
            DEBATES_DIR,
            '%s_%s_%s.json' % ('debate', self.chamber, self.date)
        )

    @property
    def content_by_speaker(self):
        self.load_data()
        speakers = defaultdict(list)
        for section in self.debate_sections:
            for speech in section.speeches:
                if speech.member_obj is not None:
                    speakers[speech.member_obj.pid].extend(speech.paras)
                else:
                    # Couldn't get pid but want to include anyway
                    speakers[speech.by].extend(speech.paras)
        return speakers

    @property
    def timestamp(self):
        self.load_data()
        return datetime.datetime.strptime(
            self.date,
            '%Y-%m-%d'
        )



from oireachtas_data.utils import get_debates


def main():
    debates = get_debates()

    print('Loaded %s debates' % len(debates))
    print('Do something with the data in the variable \'debates\'...')

    import pdb; pdb.set_trace()

    pass


if __name__ == '__main__':
    main()



import os
import pkg_resources

OIREACHTAS_DIR = os.path.join(
    '/opt/oireachtas_data',
    pkg_resources.require('oireachtas_data')[0].version.split('.')[0]
)

DEBATES_DIR = os.path.join(
    OIREACHTAS_DIR,
    'debates'
)

MEMBERS_DIR = os.path.join(
    OIREACHTAS_DIR,
    'members'
)
